package net.corda.nodeapi.internal.persistence

import com.fasterxml.jackson.databind.ObjectMapper
import liquibase.Contexts
import liquibase.LabelExpression
import liquibase.Liquibase
import liquibase.database.Database
import liquibase.database.DatabaseFactory
import liquibase.database.core.MSSQLDatabase
import liquibase.database.jvm.JdbcConnection
import liquibase.lockservice.LockServiceFactory
import liquibase.resource.ClassLoaderResourceAccessor
import net.corda.nodeapi.internal.MigrationHelpers.getMigrationResource
import net.corda.core.schemas.MappedSchema
import net.corda.core.utilities.contextLogger
import java.io.ByteArrayInputStream
import java.io.InputStream
import java.io.Writer
import javax.sql.DataSource
import liquibase.database.core.PostgresDatabase
import liquibase.structure.DatabaseObject
import liquibase.structure.core.Schema

class SchemaMigration(
        val schemas: Set<MappedSchema>,
        val dataSource: DataSource,
        private val databaseConfig: DatabaseConfig,
        private val classLoader: ClassLoader = Thread.currentThread().contextClassLoader) {

    companion object {
        private val logger = contextLogger()
    }

    /**
     * Main entry point to the schema migration.
     * Called during node startup.
     */
    fun nodeStartup(existingCheckpoints: Boolean, isH2Database: Boolean) {
        when {
            databaseConfig.initialiseSchema && isH2Database -> {
                migrateOlderDatabaseToUseLiquibase(existingCheckpoints)
                runMigration(existingCheckpoints)
            }
            databaseConfig.initialiseSchema -> runMigration(existingCheckpoints)
            else -> checkState()
        }
    }

    /**
     * Will run the Liquibase migration on the actual database.
     */
    fun runMigration(existingCheckpoints: Boolean) = doRunMigration(run = true, outputWriter = null, check = false, existingCheckpoints = existingCheckpoints)

    /**
     * Will write the migration to a [Writer].
     */
    fun generateMigrationScript(writer: Writer) = doRunMigration(run = false, outputWriter = writer, check = false)

    /**
     * Ensures that the database is up to date with the latest migration changes.
     */
    fun checkState() = doRunMigration(run = false, outputWriter = null, check = true)

    /**
     * Can be used from an external tool to release the lock in case something went terribly wrong.
     */
    fun forceReleaseMigrationLock() {
        dataSource.connection.use { connection ->
            LockServiceFactory.getInstance().getLockService(getLiquibaseDatabase(JdbcConnection(connection))).forceReleaseLock()
        }
    }

    /**  Create a resourse accessor that aggregates the changelogs included in the schemas into one dynamic stream. */
    private class CustomResourceAccessor(val dynamicInclude: String, val changelogList: List<String?>, classLoader: ClassLoader) : ClassLoaderResourceAccessor(classLoader) {
        override fun getResourcesAsStream(path: String): Set<InputStream> {
            if (path == dynamicInclude) {
                // Create a map in Liquibase format including all migration files.
                val includeAllFiles = mapOf("databaseChangeLog" to changelogList.filter { it != null }.map { file -> mapOf("include" to mapOf("file" to file)) })

                // Transform it to json.
                val includeAllFilesJson = ObjectMapper().writeValueAsBytes(includeAllFiles)

                // Return the json as a stream.
                return setOf(ByteArrayInputStream(includeAllFilesJson))
            }
            return super.getResourcesAsStream(path)?.take(1)?.toSet() ?: emptySet()
        }
    }

    private fun doRunMigration(run: Boolean, outputWriter: Writer?, check: Boolean, existingCheckpoints: Boolean? = null) {

        // Virtual file name of the changelog that includes all schemas.
        val dynamicInclude = "master.changelog.json"

        dataSource.connection.use { connection ->

            // Collect all changelog file referenced in the included schemas.
            // For backward compatibility reasons, when failOnMigrationMissing=false, we don't manage CorDapps via Liquibase but use the hibernate hbm2ddl=update.
            val changelogList = schemas.map { mappedSchema ->
                val resource = getMigrationResource(mappedSchema, classLoader)
                when {
                    resource != null -> resource
                    else -> throw MissingMigrationException(mappedSchema)
                }
            }

            val customResourceAccessor = CustomResourceAccessor(dynamicInclude, changelogList, classLoader)

            val liquibase = Liquibase(dynamicInclude, customResourceAccessor, getLiquibaseDatabase(JdbcConnection(connection)))

            val schemaName: String? = databaseConfig.schema
            if (!schemaName.isNullOrBlank()) {
                if (liquibase.database.defaultSchemaName != schemaName) {
                    logger.debug("defaultSchemaName=${liquibase.database.defaultSchemaName} changed to $schemaName")
                    liquibase.database.defaultSchemaName = schemaName
                }
                if (liquibase.database.liquibaseSchemaName != schemaName) {
                    logger.debug("liquibaseSchemaName=${liquibase.database.liquibaseSchemaName} changed to $schemaName")
                    liquibase.database.liquibaseSchemaName = schemaName
                }
            }
            logger.info("defaultSchemaName=${liquibase.database.defaultSchemaName}")
            logger.info("liquibaseSchemaName=${liquibase.database.liquibaseSchemaName}")
            logger.info("outputDefaultSchema=${liquibase.database.outputDefaultSchema}")

            val unRunChanges = liquibase.listUnrunChangeSets(Contexts(), LabelExpression())

            when {
                (run && !check) && (unRunChanges.isNotEmpty() && existingCheckpoints!!) -> throw CheckpointsException() // Do not allow database migration when there are checkpoints
                run && !check -> liquibase.update(Contexts())
                check && !run && unRunChanges.isNotEmpty() -> throw OutstandingDatabaseChangesException(unRunChanges.size)
                check && !run -> {} // Do nothing will be interpreted as "check succeeded"
                (outputWriter != null) && !check && !run -> liquibase.update(Contexts(), outputWriter)
                else -> throw IllegalStateException("Invalid usage.")
            }
        }
    }

    private fun getLiquibaseDatabase(conn: JdbcConnection): Database {

        // Enterprise only
        // The standard MSSQLDatabase in Liquibase does not support sequences for Ms Azure.
        // this class just overrides that behaviour
        class AzureDatabase(conn: JdbcConnection) : MSSQLDatabase() {
            init {
                this.connection = conn
            }

            override fun getShortName(): String = "azure"

            override fun supportsSequences(): Boolean = true
        }

        // Enterprise only
        // Postgres - Wrap schema name into double quotes to preserve case sensitivity
        class PostgresDatabaseFixed(conn: JdbcConnection) : PostgresDatabase() {
            init {
                this.connection = conn
            }

            override fun getPriority(): Int { return super.getPriority() + 1 }

            // For PostgreSQL if a schema name has uppercase or lowercase characters only then Liquibase would add it to generated query
            // without double quotes and effectively make them lowercase. This is inconsistent with Corda which wraps schema name for PostgreSQL in double quotes.
            // The overridden method ensures Liquibase wraps schema name into double quotes.
            override fun mustQuoteObjectName(objectName: String, objectType: Class<out DatabaseObject>?): Boolean {
                return if (objectType == Schema::class.java)
                    true
                else
                    super.mustQuoteObjectName(objectName, objectType)
            }
        }

        val liquibaseDbImplementation = DatabaseFactory.getInstance().findCorrectDatabaseImplementation(conn)

        return when (liquibaseDbImplementation) {
            is PostgresDatabase -> PostgresDatabaseFixed(conn) // Enterprise only
            is MSSQLDatabase -> AzureDatabase(conn) // Enterprise only
            else -> liquibaseDbImplementation
        }
    }

    /** For existing database created before verions 4.0 add Liquibase support - creates DATABASECHANGELOG and DATABASECHANGELOGLOCK tables and mark changesets are executed. */
    private fun migrateOlderDatabaseToUseLiquibase(existingCheckpoints: Boolean): Boolean {
        //workaround to detect that if Corda finance module is in use then the most recent version with Liquibase migration scripts was deployed
        if (schemas.any { schema ->
                    (schema::class.qualifiedName == "net.corda.finance.schemas.CashSchemaV1" || schema::class.qualifiedName == "net.corda.finance.schemas.CommercialPaperSchemaV1")
                            && schema.migrationResource == null
                })
            throw DatabaseMigrationException("Detected incompatible corda-finance cordapp without database migration scripts, replace the existing corda-finance-VERSION.jar with the latest one.")

        val isExistingDBWithoutLiquibase = dataSource.connection.use {
            (it.metaData.getTables(null, null, "NODE%", null).next() &&
                    !it.metaData.getTables(null, null, "DATABASECHANGELOG%", null).next())
        }
        when {
            isExistingDBWithoutLiquibase && existingCheckpoints -> throw CheckpointsException()
            isExistingDBWithoutLiquibase -> {
                // Virtual file name of the changelog that includes all schemas.
                val dynamicInclude = "master.changelog.json"

                dataSource.connection.use { connection ->
                    // Schema migrations pre release 4.0
                    val preV4Baseline = mutableListOf("migration/common.changelog-init.xml",
                            "migration/node-info.changelog-init.xml",
                            "migration/node-info.changelog-v1.xml",
                            "migration/node-info.changelog-v2.xml",
                            "migration/node-core.changelog-init.xml",
                            "migration/node-core.changelog-v3.xml",
                            "migration/node-core.changelog-v4.xml",
                            "migration/node-core.changelog-v5.xml",
                            "migration/node-core.changelog-pkey.xml",
                            "migration/vault-schema.changelog-init.xml",
                            "migration/vault-schema.changelog-v3.xml",
                            "migration/vault-schema.changelog-v4.xml",
                            "migration/vault-schema.changelog-pkey.xml")

                    if (schemas.any { schema -> schema.migrationResource == "cash.changelog-master" })
                        preV4Baseline.addAll(listOf("migration/cash.changelog-init.xml",
                                "migration/cash.changelog-v1.xml"))

                    if (schemas.any { schema -> schema.migrationResource == "commercial-paper.changelog-master" })
                        preV4Baseline.addAll(listOf("migration/commercial-paper.changelog-init.xml",
                                "migration/commercial-paper.changelog-v1.xml"))

                    if (schemas.any { schema -> schema.migrationResource == "node-notary.changelog-master" })
                        preV4Baseline.addAll(listOf("migration/node-notary.changelog-init.xml",
                                "migration/node-notary.changelog-v1.xml"))

                    val customResourceAccessor = CustomResourceAccessor(dynamicInclude, preV4Baseline, classLoader)
                    val liquibase = Liquibase(dynamicInclude, customResourceAccessor, getLiquibaseDatabase(JdbcConnection(connection)))
                    liquibase.changeLogSync(Contexts(), LabelExpression())
                }
            }
        }
        return isExistingDBWithoutLiquibase
    }
}

open class DatabaseMigrationException(message: String) : IllegalArgumentException(message) {
    override val message: String = super.message!!
}

class MissingMigrationException(@Suppress("MemberVisibilityCanBePrivate") val mappedSchema: MappedSchema) : DatabaseMigrationException(errorMessageFor(mappedSchema)) {
    internal companion object {
        fun errorMessageFor(mappedSchema: MappedSchema): String = "No migration defined for schema: ${mappedSchema.name} v${mappedSchema.version}"
    }
}

class OutstandingDatabaseChangesException(@Suppress("MemberVisibilityCanBePrivate") private val count: Int) : DatabaseMigrationException(errorMessageFor(count)) {
    internal companion object {
        fun errorMessageFor(count: Int): String = "There are $count outstanding database changes that need to be run. Please use the advanced migration tool. See: https://docs.corda.r3.com/database-management.html"
    }
}

class CheckpointsException : DatabaseMigrationException("Attempting to update the database while there are flows in flight. " +
        "This is dangerous because the node might not be able to restore the flows correctly and could consequently fail. " +
        "Updating the database would make reverting to the previous version more difficult. " +
        "Please drain your node first. See: https://docs.corda.net/upgrading-cordapps.html#flow-drains")

class DatabaseIncompatibleException(@Suppress("MemberVisibilityCanBePrivate") private val reason: String) : DatabaseMigrationException(errorMessageFor(reason)) {
    internal companion object {
        fun errorMessageFor(reason: String): String = "Incompatible database schema version detected, please run the node with configuration option database.initialiseSchema=true. Reason: $reason"
    }
}